<?php
// $Id: deadwood.conversions.inc,v 1.2 2008/07/16 17:12:28 solotandem Exp $

/**
 * @file
 * Generate version upgrade code from 5.x to 6.x.
 *
 * Copyright 2008 by Jim Berry ("solotandem", http://drupal.org/user/240748)
 */

/**
 * Convert files in a directory.
 *
 * @param string $dirname Module input directory containing the code files to be updated.
 * @param string $newdirname Module output directory where the converted code files will be written.
 * @param array $params Parameters for the conversions to be applied.
 * @param boolean $recursive Indicates whether to recurse the subdirectories of $dirname.
 */
function deadwood_convert_dir($dirname, $newdirname, $params = array(), $recursive = TRUE) {
  static $ignore = array('.', '..', 'CVS', '.svn');

  // Create an output directory we can write to.
  if (!is_dir($newdirname)) {
    mkdir($newdirname);
  }
  else {
    deadwood_clean_directory($newdirname);
  }

  // Files to use for testing.
  $test_files = array('signup.module', 'aggregator.module', 'taxonomy.module', 'example.module');
  $params['test'] = TRUE; // For testing

  $dir = opendir($dirname);
  while ($file = readdir($dir)) {
    // This condition will be removed when module is released.
    if (array_key_exists('test', $params) && $params['test'] && in_array($file, $test_files)) {
      copy($dirname . '/' . $file, $newdirname . '/' . $file);
      drupal_set_message('Converting the file => ' . $file);
      deadwood_convert_file($newdirname . '/' . $file, $params);
    }
    elseif (!in_array($file, $ignore)) {
      if (is_dir($dirname . '/' . $file)) {
        if ($recursive) {
          deadwood_convert_dir($dirname . '/' . $file, $newdirname . '/' . $file, $params, $recursive);
        }
      }
      elseif (preg_match('@(?<!replace)(?<!settings)\.(module|inc|php)$@', $file)) {
        copy($dirname . '/' . $file, $newdirname . '/' . $file);
        drupal_set_message('Converting the file => ' . $file);
        deadwood_convert_file($newdirname . '/' . $file, $params);
      }
      else {
        copy($dirname . '/' . $file, $newdirname . '/' . $file);
      }
    }
  }
}

/**
 * Convert a file.
 *
 * @param string $filename The name of the file to convert.
 * @param array $params Keyed array with run-time parameters. Should contain
 * at least this key:
 *    'conversions' => array() // Names of the conversions to run
 */
function deadwood_convert_file($filename, $params = array()) {
  if (!file_exists($filename)) {
    return FALSE;
  }

  // Read the file and copy the contents.
  $cur = file_get_contents($filename);
  $new = $cur;

  // Extract the conversions to be applied.
  $conversions = array();
  if (isset($params) && count($params) && array_key_exists('conversions', $params)) {
    $conversions = $params['conversions'];
  }

  // Process conversions on file.
  foreach ($conversions as $conversion) {
    if (function_exists($function = 'deadwood_convert_' . $conversion)) {
      call_user_func_array($function, array(&$new));
    }
  }

  // Write the new file.
  if ($new != $cur) {
    if (!file_put_contents($filename, $new)) {
      drupal_set_message('File could not be written');
    }
    drupal_set_message('Replaced the file'); // Do we want this displayed?
  }
}

/**
 * Find the text of a particular function.
 *
 * @param string $hook The suffix of the function name to find.
 *   Example: passing $hook = 'menu' will find the function
 *   whose name ends in '_menu'.
 * @param string $file The file to search.
 * @return string The function text.
 */
function deadwood_find_hook($hook, $file) {
  $pattern = '/^function \w+_' . $hook . '\(.*?(?=^function )/ms';
  preg_match($pattern, $file, $matches);

  return isset($matches[0]) ? $matches[0] : '';
}

/**
 * Loop on from and to arrays, converting the text of the subject string.
 *
 * @param string $from The pattern to search for.
 * @param string $to The string to replace the pattern with.
 * @param string $subject The string to search and replace.
 */
function deadwood_do_conversions($from, $to, &$subject) {
  for ($i = 0; $i < count($from); $i++) {
    $subject = preg_replace($from[$i], $to[$i], $subject);
  }
}

/**
 * Save the changes back to the file.
 *
 * @param string $cur The string to search for in $file.
 * @param string $new The replacement string.
 * @param string $file The text being replaced (all of or part of a file).
 * @param string $hook The hook being modified.
 */
function deadwood_save_changes($cur, $new, &$file, $hook) {
  if ($new != $cur) {
    $file = str_replace($cur, $new, $file);
    drupal_set_message('Completed conversions for ' . $hook);
  }
}

/**
 * Convert menu function.
 * See: http://drupal.org/node/103114.
 *
 * @param string $file The file to convert.
 */
function deadwood_convert_menu(&$file) {
  $hook = 'menu';
  $cur = deadwood_find_hook($hook, $file);
  $new = $cur;

  $msg1 = "/* TODO\n" .
          "   Non menu code that was placed in hook_menu under the '!\$may_cache' block\n" .
          "   so that it could be run during initialization, should now be moved to hook_init.\n".
          "   Previously we called hook_init twice, once early in the bootstrap process, second\n" .
          "   just after the bootstrap has finished. The first instance is now called boot\n" .
          "   instead of init.\n" .
          "   \n" .
          "   Dynamic menu items under a '!\$may_cache' block can often be simplified\n" .
          "   to remove references to arg(n) and use of '%<function-name>' to check\n" .
          "   conditions. See http://drupal.org/node/103114.\n" .
          "   \n" .
          "   The title and description arguments should not have strings wrapped in t(),\n" .
          "   because translation of these happen in a later stage in the menu system.\n" .
          "*/\n";

  $from = array();
  $to = array();
  // hook_menu() no longer takes any parameters (remove $may_cache).
  $from[] = '/_menu\(\$may_cache/m';
  $to[] = '_menu(';
  $from[] = '/^(.*?if \((!|)\$may_cache\).*)$/m';
  $to[] = "$msg1$1";
  // callback becomes page callback.
  $from[] = '/\'callback\'/m';
  $to[] = '\'page callback\'';
  // callback arguments becomes page arguments.
  $from[] = '/\'callback arguments\'/m';
  $to[] = '\'page arguments\'';
  // access becomes access callback and access arguments.
  $from[] = '/\'access\' => user_access/m';
  $to[] = '\'access arguments\' => array(';
  $from[] = '/\'access\'/m';
  $to[] = '\'access arguments\'';
  $from[] = '/= user_access\(\'/m';
  $to[] = '= array(\'';
  // title and description arguments should not have strings wrapped in t().
//  $from[] = '/\'title\' => t\(\'(.*)\'\)([|,])$/m';
//  $to[] = '\'title\' => \'$1\'$2';
//  $from[] = '/\'description\' => t\(\'(.*)\'\)([|,])$/m';
//  $to[] = '\'description\' => \'$1\'$2';
  // There may be other t() references on these lines. Add a message?
  //  t\((('|)(\w+)('|))\)
//  $from[] = '/(\'title\' =>.*\W)t\(((\'|)([\w ]+)(\'|))\)/';
  $from[] = '/(\'title\' =>.*\W)t\((.*)\)/';
  $to[] = '$1$2';
  $from[] = '/(\'description\' =>.*\W)t\((.*)\)/';
  $to[] = '$1$2';
  // Repeat these as they may occur multiple times on a line (probably a way to do this with one call).
//  $from[] = '/(\'title\' =>.*\W)t\(((\'|)([\w ]+)(\'|))\)/';
  $from[] = '/(\'title\' =>.*\W)t\((.*)\)/';
  $to[] = '$1$2';
  $from[] = '/(\'description\' =>.*)\Wt\((.*)\)/';
  $to[] = '$1$2';
  // Menu index strings with wildcards may be simplified
  $from[] = '/(\'path\' =>.*)arg\(\d+\)(.*)$/m'; // '/(\'path\' =>.*)\..*arg\(\d+\).*\.(.*)$/m';
  $to[] = '$1\'%\'$2';

  deadwood_do_conversions($from, $to, $new);
  deadwood_save_changes($cur, $new, $file, $hook);
}

/**
 * Convert calls to url() and l().
 *
 * See: http://drupal.org/node/114774#url.
 * Adapted from the script at http://drupal.org/files/issues/replace.php_.txt.
 *
 * @param string $file The file to convert.
 */
function deadwood_convert_url(&$file) {
  $hook = 'url';
  $msg = "/* TODO\n" .
         "   Please manually fix the parameters on the l() or url() function on the next line.\n" .
         "   Typically, this was not changed because of a function call inside an array call like\n" .
         "   array('title' => t('View user profile.'))." .
         "*/\n";

  $callbacks = array('deadwwod_fix_l', 'deadwwod_fix_url');
  foreach ($callbacks as $callback) {
    $cur = $file;
    $new = '';

    // Process file in chunks to avoid PCRE bug?
    $pattern = $callback == 'deadwwod_fix_l' ? '/(?=(?<!ion )(?<![a-zA-Z_\'\"])l\()/' : '/(?=(?<!ion )(?<![a-zA-Z_\'\"])url\()/';
    $chunks = preg_split($pattern, $cur);

    $search = $callback == 'deadwwod_fix_l' ? '@(?<!ion )(?<![a-zA-Z_\'\"])l\((([^()]*|\((?2)?\))+)\)@' : '@(?<!ion )(?<![a-zA-Z_\'\"])url\((([^()]*|\((?2)?\))+)\)@';

    foreach ($chunks as $chunk) {
      $chg = preg_replace_callback($search, $callback, $chunk);
      $new .= $chg != '' ? $chg : $msg . $chunk;
    }

    deadwood_save_changes($cur, $new, $file, $hook);
  }
}

/**
 * Callback function for converting l() references.
 *
 * @param array $matches Function arguments.
 * @return Text of converted function call.
 */
function deadwwod_fix_l($matches) {
  $args = deadwwod_get_args($matches[1]);

  // Two arguments, no change.
  if (count($args) <= 2) {
    return $matches[0];
  }

  $options = "array(";
  if (isset($args[2]) && $args[2][1] != 'array()' && $args[2][1] != 'NULL') {
    $options .= "'attributes' => " . $args[2][1] . ', ';
  }
  if (isset($args[3]) && $args[3][1] != "''" && $args[3][1] != '""' && $args[3][1] != 'NULL') {
    $options .= "'query' => " . $args[3][1] . ', ';
  }
  if (isset($args[4]) && $args[4][1] != "''" && $args[4][1] != '""' && $args[4][1] != 'NULL') {
    $options .= "'fragment' => " . $args[4][1] . ', ';
  }
  if (isset($args[5]) && $args[5][1] != "FALSE") {
    $options .= "'absolute' => TRUE, ";
  }
  if (isset($args[6]) && $args[6][1] != "FALSE") {
    $options .= "'html' => TRUE, ";
  }
  $options = rtrim($options, ', ');

  $return = 'l('. $args[0][1] . ', ' . $args[1][1] . ', ' . $options  . '))';
  return $return;
}

/**
 * Callback function for converting url() references.
 *
 * @param array $matches Function arguments.
 * @return Text of converted function call.
 */
function deadwwod_fix_url($matches) {
  $args = deadwwod_get_args($matches[1]);

  // One argument, no change.
  if (count($args) <= 1) {
    return $matches[0];
  }

  // Ignore the url() call from l().
  if ($args[1][1] == '$options') return $matches[0];

  $options = "array(";
  if (isset($args[1]) && $args[1][1] != "''" && $args[1][1] != '""' && $args[1][1] != 'NULL') {
    $options .= "'query' => " . $args[1][1] . ', ';
  }
  if (isset($args[2]) && $args[2][1] != "''" && $args[2][1] != '""' && $args[2][1] != 'NULL') {
    $options .= "'fragment' => " . $args[2][1] . ', ';
  }
  if (isset($args[3]) && $args[3][1] != "FALSE") {
    $options .= "'absolute' => TRUE, ";
  }
  $options = rtrim($options, ', ');

  $return = 'url(' . $args[0][1] . ', ' . $options  . '))';
  return $return;
}

/**
 * Helper function for parsing arguments.
 *
 * @param array $matches Function arguments.
 * @return Text of converted function call.
 */
function deadwwod_get_args($string) {
  preg_match_all('` *(([^(),]+|\((([^()]*|\((?3)?\))+)\))+) *(,|$)`', $string, $matches, PREG_SET_ORDER);
  return $matches;
}

/**
 * Convert calls to taxonomy_node_get_terms(), taxonomy_node_get_terms_by_vocabulary(),
 * taxonomy_node_save and taxonomy_node_delete().
 *
 * @param string $file The file to convert.
 */
function deadwood_convert_taxonomy(&$file) {
  $hook = 'taxonomy';
  $cur = $file;
  $new = $cur;

  $from = array();
  $to = array();
  /*
   * TODO Add this text if table name is found.
   * The {term_node} table now has a vid (version id) column, and (tid, nid, vid) is now the primary key.
   * Any direct queries against {term_node} should be checked, and probably modified to use the node
   * revision (vid) field instead of nid.
   *
  */
  // taxonomy_node_get_terms() now takes a full $node object, not just a nid (node id).
  $from[] = '/taxonomy_node_get_terms\s*\(\$(nid|node->nid)/';
  $to[] = 'taxonomy_node_get_terms($node';
//  $from[] = '/taxonomy_node_get_terms\(([^\$node]+)/m';
//  $to[] = '_menu(';
  // taxonomy_node_get_terms_by_vocabulary() now takes a full $node object, not just a nid (node id).
  $from[] = '/taxonomy_node_get_terms_by_vocabulary\s*\(\$(nid|node->nid)/m';
  $to[] = 'taxonomy_node_get_terms_by_vocabulary($node';
  // taxonomy_node_save() now takes a full $node object, not just a nid (node id).
  $from[] = '/taxonomy_node_save\s*\(\$(nid|node->nid)/m';
  $to[] = 'taxonomy_node_save($node';
  // taxonomy_node_delete() now takes a full $node object, not just a nid (node id).
  $from[] = '/taxonomy_node_delete\s*\(\$(nid|node->nid)/m';
  $to[] = 'taxonomy_node_delete($node';
  // taxonomy_node_delete_revision() has been added to delete all taxonomy terms from the current revision of the given $node object.
  
  deadwood_do_conversions($from, $to, $new);
  deadwood_save_changes($cur, $new, $file, $hook);
}

/**
 * Convert calls to format_plural() to allow for variable replacements like t().
 *
 * @param string $file The file to convert.
 */
function deadwood_convert_format_plural(&$file) {
  $hook = 'format_plural';
  $cur = $file;
  $new = $cur;

  $msg = "/* TODO\n" .
         "   An argument for replacements has been added to format_plural(),\n" .
         "   escaping and/or theming the values just as done with t()." .
         "*/";

  $from = array();
  $to = array();
  // format_plural() accepts replacements.
  /* 5.x code:
   * strtr(format_plural($num, 'There is currently 1 %type post on your site.', 'There are currently @count %type posts on your site.'), array('%type' => theme('placeholder', $type)));
   * becomes 6.x code:
   * format_plural($num, 'There is currently 1 %type post on your site.', 'There are currently @count %type posts on your site.', array('%type' => $type));
   */
//  $from[] = '/[\s(]strtr\s*\(format_plural\s*\(.*\),\s*array\(.*=>\s*theme\(.*,(\w)*\)/';
//  $to[] = $msg . '$1'; // We could place the msg at the top of the file with other messages???
//  $from[] = '/format_plural\s*\(/';
//  $to[] = $msg . '$1';
  // Find the Id line in the file.
  $from[] = '/^(\/\/ \$Id.*\$$)/m';
  $to[] = "$1\n\n$msg";
  
  deadwood_do_conversions($from, $to, $new);
  deadwood_save_changes($cur, $new, $file, $hook);
}

/**
 * Convert calls to hook_form_alter() for parameter changes.
 *
 * @param string $file The file to convert.
 */
function deadwood_convert_form_alter(&$file) {
  /*
   * TODO These module functions could have added additional parameters from the core signature.
   * We need to append any additional parameters to the end of the required parameters.
   * Should we search for the string '_form_alter' and place a notification message?
   * This might be present if the module called the hook itself.
   * Are there variables inside this function we could/should rename?
   */
  $hook = 'form_alter';
  $pattern = '/(?<!hook)_form_alter\s*(\(.*\))/';
  // No need for preg_match_all as a module should only define this function once?
  if (!preg_match($pattern, $file, $matches)) {
    return;
  }
  $cur = $matches[0];
  $new = $cur;

  $search = $matches[1];
  $replace = '(&$form, $form_state, $form_id)';
  $new = str_replace($search, $replace, $new);

  deadwood_save_changes($cur, $new, $file, $hook);
}

/**
 * Convert calls to hook_link_alter() for parameter changes.
 *
 * @param string $file The file to convert.
 */
function deadwood_convert_link_alter(&$file) {
  /*
   * TODO These module functions could have added additional parameters from the core signature.
   * We need to append any additional parameters to the end of the required parameters.
   * Should we search for the string '_link_alter' and place a notification message?
   * This might be present if the module called the hook itself.
   */
  $hook = 'link_alter';
  $pattern = '/_link_alter\s*(\(.*\))/';
  // No need for preg_match_all as a module should only define this function once?
  if (!preg_match($pattern, $file, $matches)) {
    return;
  }
  $cur = $matches[0];
  $new = $cur;

  $search = $matches[1];
  $replace = '(&$links, $node)';
  $new = str_replace($search, $replace, $new);

  deadwood_save_changes($cur, $new, $file, $hook);
}

/**
 * Convert calls to hook_profile_alter() for parameter changes.
 *
 * @param string $file The file to convert.
 */
function deadwood_convert_profile_alter(&$file) {
  /*
   * TODO These module functions could have added additional parameters from the core signature.
   * We need to append any additional parameters to the end of the required parameters.
   * Should we search for the string '_profile_alter' and place a notification message?
   * This might be present if the module called the hook itself.
   */
  $hook = 'profile_alter';
  $cur = deadwood_find_hook($hook, $file);
  $new = $cur;

  $from = array();
  $to = array();
  // Change function paramters.
  $from[] = '/_profile_alter\s*(\(.*\))/';
  $to[] = '_profile_alter(&$account)';
  // Change variable references.
  $from[] = '/\$fields\[/';
  $to[] = '$account->content[';

  deadwood_do_conversions($from, $to, $new);
  deadwood_save_changes($cur, $new, $file, $hook);
}

/**
 * Convert calls to hook_mail_alter() for parameter changes.
 *
 * @param string $file The file to convert.
 */
function deadwood_convert_mail_alter(&$file) {
  /*
   * TODO These module functions could have added additional parameters from the core signature.
   * We need to append any additional parameters to the end of the required parameters.
   * Should we search for the string '_mail_alter' and place a notification message?
   * This might be present if the module called the hook itself.
   */
  $hook = 'mail_alter';
  $cur = deadwood_find_hook($hook, $file);
  $new = $cur;

  $from = array();
  $to = array();
  // Change function paramters.
  $from[] = '/_mail_alter\s*(\(.*\))/';
  $to[] = '_mail_alter(&$message)';
  // Change variable references.
  $from[] = '/([^&])\$(mailkey|to|subject|body|from|headers)/';
  $to[] = '$1$message[\'$2\']';

  deadwood_do_conversions($from, $to, $new);
  deadwood_save_changes($cur, $new, $file, $hook);
}

/**
 * Add note about need to register theme functions in hook_theme().
 *
 * @param string $file The file to convert.
 */
function deadwood_convert_theme(&$file) {
  $hook = 'theme';
  $cur = $file;
  $new = $cur;

  $msg =
"/* TODO Implement the hook_theme registry. */
function something_theme() {
  return array(
    'something_format' => array(
      'file'      => 'something.inc',
      'template'  => 'something-format',
      'arguments' => array('content' => NULL),
    ),
  );
}";

  $from = array();
  $to = array();
  // Find the Id line in the file.
  $from[] = '/^(\/\/ \$Id.*\$$)/m';
  $to[] = "$1\n\n$msg";

  deadwood_do_conversions($from, $to, $new);
  deadwood_save_changes($cur, $new, $file, $hook);
}

/**
 * Add note about node/add automatic menu generation and alteration.
 *
 * @param string $file The file to convert.
 */
function deadwood_convert_node_add(&$file) {
  $hook = 'menu';
  $cur = deadwood_find_hook($hook, $file);
  $new = $cur;

  $msg =
"/* TODO
   The node/add/\$type menu items are now auto-generated by the menu system. You should not declare them in your menu hook.
   This means that you can use hook_menu_alter to change the visibility of an item or change the access callback. */";

  $from = array();
  $to = array();
  // node/add is now menu generated
  $from[] = '/^(.*\'path\'\s*=>\s*\'node\/add)/m';
  $to[] = "\n$msg\n\n$1";

  deadwood_do_conversions($from, $to, $new);
  deadwood_save_changes($cur, $new, $file, $hook);
}

/**
 * Add note about new hook_watchdog.
 *
 * @param string $file The file to convert.
 */
function deadwood_convert_hook_watchdog(&$file) {
  $hook = 'hook_watchdog';
  $cur = $file;
  $new = $cur;

  $msg =
"/* TODO
   There is now a new hook_watchdog in core. This means that contributed modules can implement hook_watchdog to log Drupal events to custom destinations. Two core modules are included, dblog.module (formerly known as watchdog.module), and syslog.module. Other modules in contrib include an emaillog.module, included in the logging_alerts module. See syslog or emaillog for an example on how to implement hook_watchdog. */
function example_watchdog(\$log = array()) {
    if (\$log['severity'] == WATCHDOG_ALERT) {
      mysms_send(\$log['user']->uid,
        \$log['type'],
        \$log['message'],
        \$log['variables'],
        \$log['severity'],
        \$log['referer'],
        \$log['ip'],
        format_date(\$log['timestamp'])));
}";

  // Find the Id line in the file.
  $from[] = '/^(\/\/ \$Id.*\$$)/m';
  $to[] = "$1\n\n$msg";

  deadwood_do_conversions($from, $to, $new);
  deadwood_save_changes($cur, $new, $file, $hook);
}

/**
 * Convert calls to watchdog() for parameter changes.
 *
 * @param string $file The file to convert.
 */
function deadwood_convert_watchdog(&$file) {
  $hook = 'watchdog';
  $cur = $file;
  $new = $cur;

  $from = array();
  $to = array();
  // Parameters of watchdog() changed
  // remove the t(.., array(.., function())) reference inside the watchdog() call
  $from[] = '/(watchdog\(.*),\s*t\((.*array.*?\)\))\)/';
  $to[] = "\$1, $2";
  // remove the t(.., array()) reference inside the watchdog() call
  $from[] = '/(watchdog\(.*),\s*t\((.*array.*?\))\)/';
  $to[] = "\$1, $2";
  // remove the t() reference inside the watchdog() call
  $from[] = '/(watchdog\(.*),\s*t\((.*?)\)/';
  $to[] = "\$1, $2";

  deadwood_do_conversions($from, $to, $new);
  deadwood_save_changes($cur, $new, $file, $hook);
}
